/*
 * Copyright 2009-2013 IPOL Image Processing On Line http://www.ipol.im/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file libpansharpening.cpp
 * @brief Functions which simulates the panshapening problem and solves it with
 * the nonlocal algorithm provided by Buades et al. [1].
 * @author Joan Duran <joan.duran@uib.es>
 */

#include "libpansharpening.h"

/**
 * \brief  Compute grayscale image of an RGB image.
 *
 * @param[in]  color  RGB image: the first pointer accounts for the channel and
 *             the second one for the pixel position.
 * @param[out] pan  panchromatic grayscale image: the pointer accounts for the
 *             pixel position.
 * @param[in]  alpha  mixing coefficients that give the panchormatic image as
 *             a linear combination of the spectral channels: the pointer
 *             accounts for the channel.
 * @param[in]  num_channels  number of channels of the image.
 * @param[in]  width, height  image size.
 * @return 1 if exit success.
 *
 */

int panchromatic(float **color, float *pan, float *alpha, int num_channels,
                 int width, int height)
{
    int dim = width * height;
    fpClear(pan, 0.0f, dim);

    // Compute grayscale image from RGB truth image
    for(int i = 0; i < dim; i++)
        for(int c = 0; c < num_channels; c++)
            pan[i] += alpha[c] * color[c][i];

    return 1;
}

/**
 * \brief  Compute low-resolution RGB image from the high-resolution one by
 *         a convolution with Gaussian kernel followed by subsampling.
 *
 * @param[in]   sampling_factor  sampling factor.
 * @param[in]   std  standard deviation of the convolution Gaussian kernel.
 * @param[in]   truth  high-resolution RGB image: the first pointer accounts
 *              for the channel and the second one for the pixel position.
 * @param[in]   num_channels  number of channels of the image.
 * @param[in]   high_width, high_height  truth image size.
 * @param[out]  low_width, low_height  low-resolution image size.
 * @return  low-resolution RGB image.
 * 
 */  

float **low_resolution(int sampling_factor, double std, float **truth,
                       int num_channels, int high_width, int high_height,
                       int & low_width, int & low_height)
{
    int high_dim = high_width * high_height;

    float **convol = new float*[num_channels];
    for(int c = 0; c < num_channels; c++)
        convol[c] = new float[high_dim];

    // Convolution by Gaussian kernel
    for(int c = 0; c < num_channels; c++)
        fiGaussianConvol(truth[c], convol[c], high_width, high_height,
                         (float) std);

    // Subsampling
    float **lowspectral = new float*[num_channels];

    for(int c = 0; c < num_channels; c++)
        lowspectral[c] = fiImageSample(convol[c], sampling_factor, high_width,
                                      high_height, low_width, low_height);

    // Free memory
    for(int c = 0; c < num_channels; c++)
        delete[] convol[c];
    delete[] convol;

    return lowspectral;
}

/**
 * \brief  Extend the low-resolution image to the high-resolution by a simple
 *         replication of a given sampling factor.
 *
 * @param[in]  sampling_factor  sampling factor.
 * @param[in]  lowres  subsampled low-resolution RGB image: the first pointer
 *             accounts for the channel and the second one for the pixel
 *             position.
 * @param[out] extended  extended image: the first pointer accounts for the
 *             channel and the second one for the pixel position.
 * @param[in]  num_channels  number of channels of the image.
 * @param[in]  high_width, high_height  extended image size.
 * @param[in]  low_width  low-resolution image size.
 * @return 1 if exit success.
 *
 */

int extension(int sampling_factor, float **lowres, float **extended,
              int num_channels, int high_width, int high_height,
              int low_width)
{
    for(int c = 0; c < num_channels; c++)
        for(int jp = 0; jp < high_height; jp++)
        {
            int js = jp / sampling_factor;

            for(int ip = 0; ip < high_width; ip++)
            {
                int is = ip / sampling_factor;		
   
                extended[c][jp*high_width+ip] = lowres[c][js*low_width+is];
            }
        }

    return 1;
}

/**
 * \brief  Compute pansharpening image through interpolation by splines.
 *
 * @param[in]  sampling_factor  sampling factor.
 * @param[in]  lowresolution  low-resolution subsampled RGB image: the first
 *             pointer accounts for the channel and the second one for the pixel
 *             position.
 * @param[out] splines  pansharpened image by splines: the first pointer
 *             accounts for the channel and the second one for the pixel
 *             position.
 * @param[in]  num_channels  number of channels of the image.
 * @param[in]  high_width, high_height  pansharpened image size.
 * @param[in]  low_width, low_height  low-resolution image size.
 * @return 1 if exit success.
 *
 */

int init_splines(int sampling_factor, float **lowresolution, float **splines,
                  int num_channels, int high_width, int high_height,
                  int low_width, int low_height)
{
    // Homography
    double **H = new double*[3];
    for(int c = 0; c < 3; c++)
        H[c] = new double[3];

    H[0][0] = (double) sampling_factor; H[0][1] = 0.0f; H[0][2] = 0.0f;
    H[1][0] = 0.0f; H[1][1] = (double) sampling_factor; H[1][2] = 0.0f;
    H[2][0] = 0.0f; H[2][1] = 0.0f; H[2][2] = 1.0f;

    // Interpolation by cubic splines
    for(int c = 0; c < num_channels; c++)
        apply_homography(lowresolution[c], low_width, low_height, H, splines[c],
                         high_width, high_height);

    // Free memory
    for(int c = 0; c<3; c++)
        delete[] H[c];
    delete[] H;

    return 1;
}

/**
 * \brief  Compute pansharpened image through IHS transform.
 *
 * @param[in]  init  initial image: the first pointer accounts for the channel 
 *             and the second one for the pixel position.
 * @param[out] fused  pansharpened image: the first pointer accounts for the
 *             channel and the second one for the pixel position.
 * @param[in]  pan  panchromatic image: the pointer accounts for the pixel
 *             position.
 * @param[in]  num_channels  number of channels of the image.
 * @param[in]  pan_width, pan_height  panchromatic image size.
 * @return 1 if exit success.
 *
 */    

int ihs(float **init, float **fused, float *pan, int num_channels, 
         int pan_width, int pan_height)
{

    int pan_dim = pan_width * pan_height;

    // Compute intensity image	
    float *intensity = new float[pan_dim];
    fpClear(intensity, 0.0f, pan_dim);
    
    for(int i = 0; i < pan_dim; i++)
        for(int c = 0; c < num_channels; c++)
            intensity[i] += 0.333333f * init[c][i];
        
    // Compute pansharpened image
    for(int c = 0; c < num_channels; c++)
        for(int i = 0; i < pan_dim; i++)
            fused[c][i] = init[c][i] + pan[i] - intensity[i];

    // Free memory
    delete[] intensity;

    return 1;
}

/**
 * \brief  Compute the weights of the regularization term. For each pixel,
 * compute the distance between the current pixel and each neighbouring pixel
 * on the panchromatic image.
 *
 * @param[in]  pan  panchromatic image: the pointer accounts for the pixel
 *             position.
 * @param[out] weights  weigths of regularization term: the first pointer
 *             accounts for the central pixel position and the second one for
 *             each neighbouring pixel position.
 * @param[in]  h  filtering parameter that controls the decay of the weights.
 * @param[in]  reswind  half window size for searching similar pixels.
 * @param[in]  compwind  half window size for computing similarity between
 *             neighbouring pixels.
 * @param[in]  pan_width, pan_height  panchromatic image size.
 *
 */

void compute_weights(float *pan, float **weights, float h, int reswind,
                     int compwind, int pan_width, int pan_height)
{
    // Parameters
    int resdim = (2 * reswind + 1) * (2 * reswind + 1);
    int compside = 2 * compwind + 1;
    h = compside * compside * h * h;
    
    // Tabulate the function Exp(-x) for x>0.
    int luttaille = (int) (LUTMAX * LUTPRECISION);
    float *lut = new float[luttaille];
    wxFillExpLut(lut, luttaille);
    
    // Compute weights on panchromatic by NL-means algorithm
    for(int y = compwind; y < pan_height - compwind; y++)
    {
        for(int x = compwind; x < pan_width - compwind; x++)
        {
            // Index of current pixel
            int l = y * pan_width + x;
            
            // Clear output
            fpClear(weights[l], 0.0f, resdim);
            
            // Learning zone depending on the window size
            int imin = MAX(x - reswind, compwind);
            int jmin = MAX(y - reswind, compwind);
            int imax = MIN(x + reswind, pan_width - compwind - 1);
            int jmax = MIN(y + reswind, pan_height - compwind - 1);
            
            // Auxiliar variables
            float weight_max = 0.0f;
            float weight_sum = 0.0f;
            int w = 0;
            int w_central = 0;
            
            // Compute distance for each pixel in the resdim-neighborhood
            for(int j = jmin; j <= jmax; j++)
            {
                for(int i = imin; i <= imax; i++)
                {
                    if(i!=x || j!=y)
                    {
                        // Compute weight
                        float some = fiL2FloatDist(pan, pan, x, y, i, j,
                                                   compwind, compwind, pan_width,
                                                   pan_width);
                        some = some / h;
                        float current_weight = wxSLUT(some, lut);
                        
                        // Saving weight
                        weights[l][w] = current_weight;
                        weight_sum += current_weight;
                        
                        if (current_weight > weight_max)
                            weight_max = current_weight;
                        
                        w++;
                    } else
                    {
                        w_central = w;
                        w++;
                    }
                }
            }
            
            // Assign maximum weight to central pixel
            weights[l][w_central] = weight_max;
            weight_sum += weight_max;
            
            // Normalize weights
            if(weight_sum > 0.00000001f)
                for(int k = 0; k < w; k++)
                    weights[l][k] /= weight_sum;
        }
    }
    
    // Free memory
    delete[] lut;
}

/**
 * \brief  Compute the spectral constraint. Compute the part of the numerical
 * scheme associated to the spectral constraint under the assumption that the
 * low-resolution pixels are formed from the high-resolution ones by a low-pass
 * filtering (convolution with a Gaussian kernel) followed by subsampling
 * (determined by the Dirac's comb).
 *
 * @param[in]  std  standard deviation of the convolution Gaussian kernel.
 * @param[in]  extended  arbitrary extension from low-resolution to
 *             high-resolution discrete grid: the first pointer accounts for
 *             the channel and the second one for the pixel position.
 * @param[in]  init  pansharpened image at current iteration: the first pointer
 *             accounts for the channel and the second one for the pixel
 *             position.
 * @param[out] convol  solution of the spectral constraint: the first pointer
 *             accounts for the channel and the second one for the pixel
 *             position.
 * @param[in]  dirac  Dirac's comb defined by the sampling grid: the pointer
 *             accounts for the pixel position.
 * @param[in]  num_channels  number of channels of the image.
 * @param[in]  pan_width, pan_height  panchromatic image size.
 *
 */

void spectral_constraint(double std, float **extended, float **init,
                         float **convol, float *dirac, int num_channels,
                         int pan_width, int pan_height)
{           
    for(int c = 0; c < num_channels; c++)
    {
        // Convolution by Gaussian kernel
        fiGaussianConvol(init[c], convol[c], pan_width, pan_height,
                         (float) std);
        
        for(int j = 0; j < pan_height; j++)
            for (int i = 0; i < pan_width; i++)
            {
                int l = j*pan_width+i;
                convol[c][l] = dirac[l] * (convol[c][l] - extended[c][l]);
            }
        
        // Convolution by transposed Gaussian kernel
        fiGaussianConvol(convol[c], convol[c], pan_width, pan_height,
                         (float) std);
    }
}

/**
 * \brief  Nonlocal variational algorithm for pansharpening image fusion.
 * Applies Algorithm 1 to minimize the nonlocal energy functional proposed by
 * Buades et al. [1]. We solve the explicit numerical scheme given in (7).
 *
 * @param[in]  sampling_factor  sampling factor.
 * @param[in]  std  standard deviation of the convolution Gaussian kernel.
 * @param[in]  h  filtering parameter that controls the decay of the weights.
 * @param[in]  lmb  trade-off parameter linked to the panchromatic constraint.
 * @param[in]  mu  trade-off parameter linked to the spectral constraint.
 * @param[in]  num_iterations  number of iterations of the explicit scheme.
 * @param[in]  reswind  half-size of the suppoirt zone used for searching
 *             similarities among pixels in the image.
 * @param[in]  compwind  half-size of the comparison window used for computing
 *             distances on the panchromatic.
 * @param[in]  alpha  mixing coefficients that give the panchormatic image as
 *             a linear combination of the spectral channels: the pointer
 *             accounts for the channel.
 * @param[in]  dt  time-step parameter of the explicit scheme.
 * @param[in]  pan  panchromatic image: the pointer accounts for the pixel
 *             position.
 * @param[in]  initial  initial image provided by IHS transform: the first
 *             pointer accounts for the channel and the second one for the pixel
 *             position.
 * @param[in]  extended  arbitrary extension from low-resolution to
 *             high-resolution discrete grid: the first pointer accounts for
 *             the channel and the second one for the pixel position.
 * @param[out] fused  pansharpened image: the first pointer accounts for the
 *             channel and the second one for the pixel position.
 * @param[in]  num_channels  number of channels of the image.
 * @param[in]  pan_width, pan_height  panchromatic image size.
 * @return 1 if exit success.
 *
 */
   
int nonlocal_pansharpening(int sampling_factor, double std, float h, float lmb,
                           float mu, int num_iterations, int reswind,
                           int compwind, float *alpha, float dt, float *pan,
                           float **initial, float **extended, float **fused,
                           int num_channels, int pan_width, int pan_height)
{
    
    // Initializations

        // Size of the panchromatic image
        int pan_dim = pan_width * pan_height;

        // Size of weight influence or research window, in practice, 7x7
        int search_dim = (2 * reswind + 1) * (2 * reswind + 1); 

        // Weight distribution computed just once
        // w(p,q) for each pixel p and q in the research window of p
        float **weights = new float*[pan_dim];
        for(int i = 0; i < pan_dim; i++)
            weights[i] = new float[search_dim];

        // Copy initialization into fused image
        for(int c = 0; c < num_channels; c++)
            fpCopy(initial[c], fused[c], pan_dim);

        // Auxiliar image for iteration purposes
        // At the very beginning, aux contains the initialization from IHS
        float **aux = new float*[num_channels];
        for(int c = 0; c < num_channels; c++)
        {
            aux[c] = new float[pan_dim];
            fpCopy(fused[c], aux[c], pan_dim);
        }

        // Convolved image during iteration        
        float **convol = new float*[num_channels];
            for(int c = 0; c < num_channels; c++)
                convol[c] = new float[pan_dim];

        // Dirac's comb defined by sampling grid        
        float *dirac = new float[pan_dim];
        fpClear(dirac, 0.0f, pan_dim);
        for(int j = 0; j < pan_height; j += sampling_factor)
            for(int i = 0; i < pan_width; i += sampling_factor)
                dirac[j*pan_width+i] = 1.0f;

        // Rescale parameter linked to spectral constraint
        mu = mu * sampling_factor * sampling_factor;

        // Iteration variable   
        int niter = 0;

    
    // Computation of weight distribution on the panchromatic image
    // The weights are computed just once
    compute_weights(pan, weights, h, reswind, compwind, pan_width, pan_height);

    
    // Iterative process
    while(niter < num_iterations)
    {
        // Computation of the spectral constraint at current iteration
        spectral_constraint(std, extended, fused, convol, dirac, num_channels,
                            pan_width, pan_height);
        
        // For each channel
        for(int c = 0; c < num_channels; c++)
        {
            // For each pixel of the image
            for(int y = compwind; y < pan_height - compwind; y++)
            {
                for(int x = compwind; x < pan_width - compwind; x++)
                {
                    // Index of current pixel
                    int l = y * pan_width + x;
                    
                    // Learning zone depending on the window size
                    int imin = MAX(x - reswind, compwind);
                    int jmin = MAX(y - reswind, compwind);
                    int imax = MIN(x + reswind, pan_width - compwind - 1);
                    int jmax = MIN(y + reswind, pan_height - compwind - 1);
                    
                    // Index in weights[l]
                    int w = 0;
                    
                    // Auxiliary variable for computing regularization term
                    float reg_term = 0.0f;
                    
                    // For each pixel in the search_dim-neighborhood
                    for(int j = jmin; j <= jmax; j++)
                        for(int i = imin; i <= imax; i++)
                        {
                            // Index of neighborhood pixel
                            int l0 = j * pan_width + i;
                            
                            // Compute weight
                            float value = fused[c][l] - fused[c][l0];
                            float weight1 = weights[l][w];
                            float weight2 = weights[l0][search_dim-1-w];
                            
                            // Add pixel to channel average
                            reg_term += value * (weight1 + weight2);
                            
                            w++;
                        }
                    
                    // Compute panchromatic constraint
                    float pan_term = 0.0f;
                    for(int k = 0; k < num_channels; k++)
                        pan_term += alpha[k] * fused[k][l];
                    pan_term -= pan[l];
                    
                    // Set value to current pixel
                    // The following formula implements the explicit scheme (7)
                    // given in the paper "Implementation of Nonlocal
                    // Pansharpening Image Fusion" by Buades et al.
                    aux[c][l] = fused[c][l] - dt * (reg_term +
                                                    lmb * alpha[c] * pan_term +
                                                    mu * convol[c][l]);
                }
            }
            
            // Update fused image
            fpCopy(aux[c], fused[c], pan_dim);
        }
        
        niter++;
    }

    
    // Free memory
    for(int c = 0; c < num_channels; c++)
    {
        delete[] aux[c]; delete[] convol[c];
    }
  
    for(int i = 0; i < pan_dim; i++)
        delete[] weights[i];

    delete[] aux;
    delete[] convol;
    delete[] weights;
    delete[] dirac;

    return 1;
}   
